@@ -10,7 +10,8 @@ from .vars import (
     SFN_TYPE,
     RUN_TASK_LAMBDA_NAME,
     CHECK_TASK_LAMBDA_NAME,
-    UPDATE_COST_LAMBDA_NAME
+    UPDATE_COST_LAMBDA_NAME,
+    S3_ENCRYT_KEY_ID,
 )
 
 
@@ -36,8 +37,6 @@ class IAM(object):
           tibanna_policy_prefix : tibanna_default_3465 / tibanna_pony_default_3465
           prefix : tibanna_ / tibanna_pony_
         """
-        # lambda names
-        self.lambda_names = [self.run_task_lambda_name, self.check_task_lambda_name, self.update_cost_lambda_name]
         if self.lambda_type:
             self.prefix = 'tibanna_' + self.lambda_type + '_'
         else:
@@ -61,6 +60,10 @@ class IAM(object):
         self.client = boto3.client('iam')
         self.iam = boto3.resource('iam')
 
+    @property
+    def lambda_names(self):
+        return [self.run_task_lambda_name, self.check_task_lambda_name, self.update_cost_lambda_name]
+
     @property
     def iam_group_name(self):
         return self.tibanna_policy_prefix
@@ -69,11 +72,17 @@ class IAM(object):
     def policy_types(self):
         return ['bucket', 'termination', 'list', 'cloudwatch', 'passrole', 'lambdainvoke',
                 'cloudwatch_metric', 'cw_dashboard', 'dynamodb', 'ec2_desc',
-                'executions', 'pricing']
+                'executions', 'pricing', 'vpc', 'kms']
 
     def policy_arn(self, policy_type):
         return 'arn:aws:iam::' + self.account_id + ':policy/' + self.policy_name(policy_type)
 
+    def kms_key_arn(self):
+        _id = S3_ENCRYT_KEY_ID
+        if not _id:
+            _id = 'invalid'  # build a dummy arn in the case that we are not using KMS
+        return 'arn:aws:kms:' + self.region + ':' + self.account_id + ':key/' + _id
+
     def policy_suffix(self, policy_type):
         suffices = {'bucket': 'bucket_access',
                     'termination': 'ec2_termination',
@@ -86,7 +95,9 @@ class IAM(object):
                     'dynamodb': 'dynamodb',
                     'ec2_desc': 'ec2_desc',
                     'pricing': 'pricing',
-                    'executions': 'executions'}
+                    'executions': 'executions',
+                    'vpc': 'vpc_access',
+                    'kms': 'kms_key_for_s3'}
         if policy_type not in suffices:
             raise Exception("policy %s must be one of %s." % (policy_type, str(self.policy_types)))
         return suffices[policy_type]
@@ -106,7 +117,9 @@ class IAM(object):
                        'dynamodb': self.policy_dynamodb,
                        'ec2_desc': self.policy_ec2_desc_policy,
                        'pricing': self.policy_pricing,
-                       'executions': self.policy_executions}
+                       'executions': self.policy_executions,
+                       'vpc': self.policy_vpc_access,
+                       'kms': self.policy_kms_access}
         if policy_type not in definitions:
             raise Exception("policy %s must be one of %s." % (policy_type, str(self.policy_types)))
         return definitions[policy_type]
@@ -134,13 +147,20 @@ class IAM(object):
             raise Exception("role_type %s must be one of %s." % (role_type, str(self.role_types)))
         return services[role_type]
 
-    def policy_arn_list_for_role(self, role_type):
-        run_task_custom_policy_types = ['list', 'cloudwatch', 'passrole', 'bucket', 'dynamodb',
-                                        'executions', 'cw_dashboard']
-        check_task_custom_policy_types = ['cloudwatch_metric', 'cloudwatch', 'bucket', 'ec2_desc',
-                                          'termination', 'dynamodb', 'pricing']
-        update_cost_custom_policy_types = ['bucket', 'executions', 'dynamodb', 'pricing']
-        arnlist = {'ec2': [self.policy_arn(_) for _ in ['bucket', 'cloudwatch_metric']] +
+    @property
+    def policy_arn_list_for_role(self):
+        """ This function specifies the permissions given to each service/lambda function. """
+        base = ['bucket']  # policies everyone needs
+        if S3_ENCRYT_KEY_ID:  # if we have a value, this key is valid and we must add perms
+            base.append('kms')
+        # returns a dictionary with role_type as keys
+        # adding vpc access to only check_task since run_task has full ec2 access
+        run_task_custom_policy_types = base + ['list', 'cloudwatch', 'passrole', 'dynamodb',
+                                               'executions', 'cw_dashboard']
+        check_task_custom_policy_types = base + ['cloudwatch_metric', 'cloudwatch', 'ec2_desc',
+                                                 'termination', 'dynamodb', 'pricing', 'vpc']
+        update_cost_custom_policy_types = base + ['executions', 'dynamodb', 'pricing', 'vpc']
+        arnlist = {'ec2': [self.policy_arn(_) for _ in base + ['cloudwatch_metric', 'ec2_desc']] +
                           ['arn:aws:iam::aws:policy/AmazonEC2ContainerRegistryReadOnly'],
                    # 'stepfunction': [self.policy_arn(_) for _ in ['lambdainvoke']],
                    'stepfunction': ['arn:aws:iam::aws:policy/service-role/AWSLambdaRole'],
@@ -148,10 +168,8 @@ class IAM(object):
                                               ['arn:aws:iam::aws:policy/AmazonEC2FullAccess'],
                    self.check_task_lambda_name: [self.policy_arn(_) for _ in check_task_custom_policy_types],
                    self.update_cost_lambda_name: [self.policy_arn(_) for _ in update_cost_custom_policy_types]}
-                   
-        if role_type not in arnlist:
-            raise Exception("role_type %s must be one of %s." % (role_type, str(self.role_types)))
-        return arnlist[role_type]
+
+        return arnlist
 
     @property
     def instance_profile_name(self):
@@ -206,6 +224,48 @@ class IAM(object):
         }
         return policy
 
+    @property
+    def policy_vpc_access(self):
+        policy = {
+            "Version": "2012-10-17",
+            "Statement": [
+                {
+                    "Effect": "Allow",
+                    "Action": [
+                        "ec2:DescribeNetworkInterfaces",
+                        "ec2:CreateNetworkInterface",
+                        "ec2:AttachNetworkInterface",
+                        "ec2:DeleteNetworkInterface"
+                    ],
+                    "Resource": "*"
+                }
+            ]
+        }
+        return policy
+
+    @property
+    def policy_kms_access(self):
+        """ Returns a dictionary representing a policy granting some limited
+            permissions needed to upload/download files to encrypted S3 buckets
+            given an S3_ENCRYPT_KEY_ID.
+        """
+        return {
+            'Version': '2012-10-17',
+            'Statement': {
+                'Effect': 'Allow',
+                'Action': [
+                    'kms:Encrypt',
+                    'kms:Decrypt',
+                    'kms:ReEncrypt*',
+                    'kms:GenerateDataKey*',
+                    'kms:DescribeKey'
+                ],
+                'Resource': [
+                    self.kms_key_arn()
+                ]
+            }
+        }
+
     @property
     def policy_list_instanceprofiles(self):
         policy_list_instanceprofiles = {
@@ -284,7 +344,9 @@ class IAM(object):
         execution_arn_prefix = 'arn:aws:states:' + self.region + ':' + self.account_id + ':stateMachine:'
         sfn_arn_prefix = 'arn:aws:states:' + self.region + ':' + self.account_id + ':execution:'
         resources = [execution_arn_prefix + self.tibanna_sfn_name,
-                    sfn_arn_prefix + self.tibanna_sfn_name + ':*']
+                    execution_arn_prefix + self.tibanna_sfn_name + '_costupdater',
+                    sfn_arn_prefix + self.tibanna_sfn_name + ':*',
+                    sfn_arn_prefix + self.tibanna_sfn_name + '_costupdater:*']
         policy = {
             "Version": "2012-10-17",
             "Statement": [
@@ -354,7 +416,8 @@ class IAM(object):
                     "Action": [
                         "dynamodb:DescribeTable",
                         "dynamodb:PutItem",
-                        "dynamodb:Query"
+                        "dynamodb:Query",
+                        "dynamodb:UpdateItem"
                     ],
                     "Resource": "arn:aws:dynamodb:" + self.region + ":" + self.account_id + ":table/" + DYNAMODB_TABLE
                 }
@@ -372,14 +435,15 @@ class IAM(object):
                     "Action": [
                         "ec2:DescribeInstances",
                         "ec2:DescribeInstanceStatus",
-                        "ec2:DescribeSpotPriceHistory"
+                        "ec2:DescribeSpotPriceHistory",
+                        "ec2:DescribeSpotInstanceRequests"
                     ],
                     "Resource": "*"
                 }
             ]
         }
         return policy
-    
+
     @property
     def policy_pricing(self):
         policy = {
@@ -449,7 +513,7 @@ class IAM(object):
         role_policy_doc = self.role_policy_document(self.role_service(role_type))
         self.create_role_robust(self.role_name(role_type), json.dumps(role_policy_doc), verbose)
         role = self.iam.Role(self.role_name(role_type))
-        for p_arn in self.policy_arn_list_for_role(role_type):
+        for p_arn in self.policy_arn_list_for_role[role_type]:
             response = role.attach_policy(PolicyArn=p_arn)
             if verbose:
                 logger.debug("response from IAM attach_policy :" + str(response))
